/*
This product contains certain software code or other information
("AT&T Software") proprietary to AT&T Corp. ("AT&T").  The AT&T
Software is provided to you "AS IS".  YOU ASSUME TOTAL RESPONSIBILITY
AND RISK FOR USE OF THE AT&T SOFTWARE.  AT&T DOES NOT MAKE, AND
EXPRESSLY DISCLAIMS, ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND
WHATSOEVER, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, WARRANTIES OF
TITLE OR NON-INFRINGEMENT OF ANY INTELLECTUAL PROPERTY RIGHTS, ANY
WARRANTIES ARISING BY USAGE OF TRADE, COURSE OF DEALING OR COURSE OF
PERFORMANCE, OR ANY WARRANTY THAT THE AT&T SOFTWARE IS "ERROR FREE" OR
WILL MEET YOUR REQUIREMENTS.

Unless you accept a license to use the AT&T Software, you shall not
reverse compile, disassemble or otherwise reverse engineer this
product to ascertain the source code for any AT&T Software.

(c) AT&T Corp. All rights reserved.  AT&T is a registered trademark of AT&T Corp.

***********************************************************************

History:

      24/11/99  - initial release by Hartmut Liefke, liefke@seas.upenn.edu
                                     Dan Suciu,      suciu@research.att.com
*/

//********************************************************************
//********************************************************************

// This module contains the standard user compressors, such as 'u', 'u8', 'di', ...

#include "stdafx.h"

// The standard signed integer compressor 'i'
SignedIntCompressor::SignedIntCompressor(Session *s, unsigned long mymindigits): UserCompressor(s)
{
   mindigits=mymindigits;
   datasize=0;contnum=1;isrejecting=1;canoverlap=1;isfixedlen=0;
}

char SignedIntCompressor::ParseString(char *str,unsigned len,char *dataptr)
{
   return ParseSignedInt(str,len,&val,&isneg);
}

void SignedIntCompressor::CompressString(char *str,unsigned len,CompressContainer *cont,char *dataptr)
{
   cont->StoreSInt32(isneg,val);
}

SignedIntUncompressor::SignedIntUncompressor(unsigned long mymindigits): UserUncompressor()
{
	Init(mymindigits);
}

SignedIntUncompressor::SignedIntUncompressor(Session *s, unsigned long mymindigits): UserUncompressor(s)
{
	Init(mymindigits);
}

void SignedIntUncompressor::Init(unsigned long mymindigits)
{
   mindigits=mymindigits;
   datasize=0;contnum=1;
}

void SignedIntUncompressor::UncompressItem(UncompressContainer *cont,char *dataptr,XMLOutput *xoutput)
{
   char isneg;
   unsigned long val=cont->LoadSInt32(&isneg);

   PrintInteger(val,isneg,mindigits,xoutput);
}

char *SignedIntCompressorFactory::GetName()         
{  
	return "i"; 
}
char *SignedIntCompressorFactory::GetDescription()  
{  
	return "Compressor for signed integers"; 
}

UserCompressor *SignedIntCompressorFactory::InstantiateCompressor(char *paramstr,int len)
{
   unsigned long mindigits=0;
   if(paramstr!=NULL)
   {
      char savechar=paramstr[len];
      paramstr[len]=0;
      mindigits=atoi(paramstr);
      paramstr[len]=savechar;
      if(mindigits<=0)
      {
         XMillException *e = new XMillException(XMILL_ERR_ARGUMENTS, "Invalid parameter '");
         e->ErrorCont(paramstr,len);
         e->ErrorCont("' for compressor 'i'!");
         throw e;
      }
   }
   return new SignedIntCompressor(session, mindigits);
}

UserUncompressor *SignedIntCompressorFactory::InstantiateUncompressor(char *paramstr,int len)
{
   unsigned long mindigits=0;
   if(paramstr!=NULL)
   {
      char savechar=paramstr[len];
      paramstr[len]=0;
      mindigits=atoi(paramstr);
      paramstr[len]=savechar;
      return new SignedIntUncompressor(session, mindigits);
   }
   else
      return &uncompressor;
}
